/*  Copyright (C) 2008 Niels Richard Hansen. 

    This file is part of StemSearch.

    StemSearch is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.

    StemSearch is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <vector>
#include <string>
#include <cmath>
#include <fstream>
#include <algorithm>
#include "genome.h"
#include "../random_numbers/random.h"

void Genome::init_alph()
{
	DNAalph1[0] = 'a'; DNAalph1[1] = 'c'; DNAalph1[2] = 'g'; DNAalph1[3] = 't'; DNAalph1[4]='-';
	DNAalph2[0] =  'A'; DNAalph2[1] = 'C'; DNAalph2[2] = 'G'; DNAalph2[3] = 'T'; DNAalph2[4]='-';
	RNAalph1[0] = 'a'; RNAalph1[1] = 'c'; RNAalph1[2] = 'g'; RNAalph1[3] = 'u'; RNAalph1[4]='-';
	RNAalph2[0] =  'A'; RNAalph2[1] = 'C'; RNAalph2[2] = 'G'; RNAalph2[3] = 'U'; RNAalph2[4]='-';
}
	

Genome::Genome(bool w)
{
	warnings=w;
	init_alph();
}

Genome::Genome(vector<long> gen)
{
	init_alph();
	IntGenome = gen;
}
	
Genome::Genome(char* file)
{
	init_alph();
	read(file);
}
	
void Genome::init(vector<long> gen)
{
	IntGenome = gen;
}
	
long Genome::size()
{
	return(IntGenome.size());
}

char Genome::DNAalph(long i)
{
	return(DNAalph1[i]);
}

char Genome::RNAalph(long i)
{
	return(RNAalph1[i]);
}

string Genome::name()
{
	return Name;
}

long Genome::operator[](long i)
{
	return(IntGenome[i-1]);
}


string Genome::DNAsegment(long i, long j)
{
	string temp;
	
	if((i >= 1) && (i <= j) && (j <= IntGenome.size()))
	{
		temp.resize(j-i+1);
	
		for(long k=0; k<=(j-i);k++)
			temp[k] = DNAalph(operator[](i+k));

	}
	
	return(temp);
}

string Genome::RNAsegment(long i, long j)
{
	string temp;
	
	if((i >= 1) && (i <= j) && (j <= IntGenome.size()))
	{
		temp.resize(j-i+1);
	
		for(long k=0; k<=(j-i);k++)
			temp[k] = RNAalph(operator[](i+k));
	}

	return(temp);
}
	

void Genome::read(char* file)
{
	char ch;
	Name = "";
	
	IntGenome.clear();
		
	ifstream data(file);
	if(data)
	{
		data.get(ch);
		if (ch == '>')
		{
			while ((data.get(ch)) && (ch != '\n'))
			{
				if(ch != ' ') Name += ch;
			}
		}
		else
		{	
			Name = file;
		}
		         
		
		while (data.get(ch))                          
		{
			if ((ch != '\n') && (ch != ' ')) 
			{     
				long temp=4;
				if((ch ==DNAalph1[0]) || (ch==DNAalph2[0])) temp=0;
				else if((ch==DNAalph1[1]) || (ch==DNAalph2[1])) temp=1;
				else if((ch==DNAalph1[2]) || (ch==DNAalph2[2])) temp=2;
				else if((ch==DNAalph1[3]) || (ch==DNAalph2[3]) || (ch==RNAalph1[3]) || (ch==RNAalph2[3])) temp=3;
				IntGenome.push_back(temp);
			}
		}
	}
}

bool Genome::read(istream &data,char stopread)
{
	char ch;
	Name = "";
	long temp;
	
	IntGenome.resize(0);
	
	if(data)
	{
		data.get(ch);
		if (ch == '>')
		{
			int k=0;
			while (data.get(ch) && (ch != '\n'))
			{
				Name += ch;
				k++;
			}
		}
		else if(ch=='@') 
		{
			return true;
		}
		else
		{	if ((ch != '\n')  && (ch != ' ')) 
			{     
				if((ch ==DNAalph1[0]) || (ch==DNAalph2[0])) temp=0;
				else if((ch==DNAalph1[1]) || (ch==DNAalph2[1])) temp=1;
				else if((ch==DNAalph1[2]) || (ch==DNAalph2[2])) temp=2;
				else if((ch==DNAalph1[3]) || (ch==DNAalph2[3]) || (ch==RNAalph1[3]) || (ch==RNAalph2[3])) temp=3;
				else {temp=4; if(warnings) cout << "Non standard nucleotide: " << ch << "\n";}
				IntGenome.push_back(temp);
			}
			Name = "NN";
		}
		
		while ((data.get(ch)) && (ch != stopread) && (ch != '@'))                          
		{
			if ((ch != '\n') && (ch != ' ')) 
			{     
				if((ch ==DNAalph1[0]) || (ch==DNAalph2[0])) temp=0;
				else if((ch==DNAalph1[1]) || (ch==DNAalph2[1])) temp=1;
				else if((ch==DNAalph1[2]) || (ch==DNAalph2[2])) temp=2;
				else if((ch==DNAalph1[3]) || (ch==DNAalph2[3]) || (ch==RNAalph1[3]) || (ch==RNAalph2[3])) temp=3;
				else {temp=4; if(warnings) cout << "Non standard nucleotide: " << ch  << "\n";}
				IntGenome.push_back(temp);
			}
		}
		
		if (ch == '>') data.unget();
	}
	
	return ((ch == '@') || (!data));
}

void Genome::matchtable(long last, long first, long sz, long b)
	{
		SegmentSize = sz;
		Base = b; 
		TableSize = (long) pow( (double) Base, (int) SegmentSize);
		long i,j;
		Positions = new vector<long>[TableSize];
		
		if (last == -1) 
		{ 
			last = IntGenome.size();
		}
		else 
		{
			last = min(last,(long) IntGenome.size());
		}
		
		
		for(i=first;i<last;i++)
		{
			Value = table_lookup(i);
			Positions[Value].push_back(i);
		}
	} 

long Genome::table_lookup(long i)
{
	long result = IntGenome[i], p = 1;
	for(long j=1;j<SegmentSize;j++)
	{   
		p = p*Base;        
		result = result + IntGenome[i+j]*p;
	}
	return(result);
}

string Genome::inv_table_lookup(long Value)
{
	string result = "";
	long p = Value, res;
	for(long j=0;j<SegmentSize;j++)
	{   
		res = p%Base;
		p = p/Base;        
		result = result+DNAalph(res);
	}
	return(result);
}

void Genome::save_table(char* filename, char* sep)
{
	long i,j;
	ofstream file(filename);

	for(i=0;i<TableSize;i++)
	{
		file << i << " " << inv_table_lookup(i) << "    ";
		for(j=0;j<Positions[i].size();j++)
			file << Positions[i][j] << sep;
		file << "\n";
	}
	file << "\n";
}        

void Genome::generate(long n, vector<double> freq, long seed)
{
    Name="Random";
	IntGenome.resize(n);
	vector<double> ac_freq(4);
	long i,k;
	double temp;

	r.setseed(seed);
	ac_freq[0] = freq[0];

	for(i=1;i<4;i++)
		ac_freq[i] = ac_freq[i-1] + freq[i];
	

	for(i=0;i<n;i++)
		{   
			temp = r.next();
			k=0;
			while(temp > ac_freq[k])
				k++;     
			IntGenome[i] = k;
		}
}

RectMatrix<double> Genome::TP()
{
	RectMatrix<double> temp(5,5),tmp(4,1);
	temp.init();
	tmp.init();
	int i,j;
	
	for(i=1;i<IntGenome.size();i++)
		temp(IntGenome[i-1],IntGenome[i])++;
	
	for(i=0;i<4;i++)
	{
		for(j=0;j<4;j++)
			tmp(i) += temp(i,j);
	}
	
	for(i=0;i<4;i++)
	{
		for(j=0;j<4;j++)
		{	cout << temp(i,j) << " ";
			temp(i,j) /= tmp(i);
		}
		
		cout << tmp(i) << "\n";
	}
	
	return temp;
}
	

void Genome::Markov_generate(long n, RectMatrix<double> freq, long seed)
{
  Name="Random";
	IntGenome.resize(n);
	RectMatrix<double> ac_freq(4,4); 
	uni r;
	long i,j,k;
	double temp;
	
	r.setseed(seed);
	for(i=0;i<4;i++)
	{
		ac_freq(i,0) = freq(i,0);
		for(j=1;j<4;j++)
			ac_freq(i,j) = ac_freq(i,j-1) + freq(i,j);
	}

	temp = r.next();
	k=0;
	while(temp > ac_freq(0,k))
		k++;     
	IntGenome[0] = k;

	for(i=1;i<n;i++)
		{   
		  temp = r.next();
			k=0;
			while(temp > ac_freq(IntGenome[i-1],k))
				k++;     
			IntGenome[i] = k;
		}

}

void Genome::warn(bool w)
{
	warnings = w;
}
